from collections.abc import Callable, Sequence
from itertools import chain

from elasticai.creator.vhdl.design.ports import Port
from elasticai.creator.vhdl.design.signal import Signal


def signal_string(name: str, direction: str, width: int | str) -> str:
    if isinstance(width, int) and width > 0:
        data_type = f"std_logic"
    else:
        data_type = f"std_logic_vector({width} - 1 downto 0)"
    return f"{name} : {direction} {data_type}"


def _port_definition_from_port(
    p: Port, signal_width_handler: Callable[[Signal], str]
) -> list[str]:
    signals = []

    signals_and_directions = chain(zip(p.incoming, "in"), zip(p.outgoing, "out"))
    for signal, direction in signals_and_directions:
        width = signal_width_handler(signal)
        signals.append(signal_string(signal.name, direction, width))
    return wrap_lines_into_port_statement(signals)


def _extract_actual_width(s: Signal) -> str:
    if s.width == 0:
        return "std_logic"
    else:
        return f"std_logic_vector({s.width} - 1 downto 0)"


def _generate_template_width(s: Signal) -> str:
    if s.width == 0:
        return "std_logic"
    else:
        return f"std_logic_vector(${s.name}_width - downto 0)"


def vhdl_port_definition(p: Port) -> list[str]:
    return _port_definition_from_port(p, _extract_actual_width)


def template_string_for_port_definition(p: Port) -> list[str]:
    return _port_definition_from_port(p, _generate_template_width)


def wrap_lines_into_port_statement(lines: Sequence[str]) -> list[str]:
    semicolon_ended_lines = (f"{line};" for line in lines[:-1])
    last_line = lines[-1]
    return list(chain(["port ("], semicolon_ended_lines, [last_line, ");"]))


def expand_to_parameters_for_port_template(p: Port) -> dict[str, str]:
    return {
        f"{s.name}_width": f"{s.width}"
        for s in filter(lambda s: s.width > 0, p.signals)
    }



from collections.abc import Iterable, Iterator, Sequence
from dataclasses import dataclass
from itertools import chain
from typing import Any, TypeVar


@dataclass(eq=True, frozen=True)
class DataFlowNode:
    sinks: tuple[str, ...]
    sources: tuple[str, ...]
    name: str

    @classmethod
    def top(cls, name: str) -> "DataFlowNode":
        return cls(
            name=name,
            sinks=("done", "y", "x_address"),
            sources=("clock", "x", "y_address", "enable"),
        )

    @classmethod
    def buffered(cls, name: str) -> "DataFlowNode":
        return cls(
            name=name,
            sinks=("clock", "x", "y_address", "enable"),
            sources=("done", "y", "x_address"),
        )

    @classmethod
    def unbuffered(cls, name: str) -> "DataFlowNode":
        return cls(name=name, sinks=("clock", "x", "enable"), sources=("y",))


@dataclass(eq=True, frozen=True)
class WiringProtocol:
    """
    A wiring protocol specifies how we search for sources that need to be connected to sinks.
    We understand the neural network as a DAG. The direction "down" refers to the direction
    of the dataflow during inference. A "sink" refers to an "in" signal in vhdl and a source
    refers to an "out" signal. Each field of the class holds a dictionary specifying compatible
    source names for sinks. For "up_sinks" we search in ancestor nodes for satisfying sources
    and for "down_sinks" we search in successor nodes.
    """

    up_sinks: dict[str, tuple[str, ...]]
    down_sinks: dict[str, tuple[str, ...]]


BASIC_WIRING = WiringProtocol(
    up_sinks={
        "clock": ("clock",),
        "enable": ("enable", "done"),
        "x": ("x", "y"),
        "y": ("x", "y"),
        "done": ("enable", "done"),
    },
    down_sinks={
        "y_address": ("y_address", "x_address"),
        "x_address": ("y_address", "x_address"),
    },
)


class AutoWiringProtocolViolation(Exception):
    pass


T = TypeVar("T")


class AutoWirer:
    """
    Please note that this implementation performs the wiring solely based on names.
    Hence, it will not check whether signals are compatible based on their width
    or data type.
    It only supports a sequence of nodes as the input graph.
    Handling nodes with more than one child requires deciding how data paths are split
    and merged.
    """

    def __init__(self) -> None:
        self._reset()

    def _check_protocol_support_violations(self, graph, top):
        for node in chain(graph, (top,)):
            for sink in node.sinks:
                if sink not in self._all_protocol_sinks():
                    raise AutoWiringProtocolViolation(f"{sink}")
            for source in node.sources:
                if source not in self._all_protocol_sources():
                    raise AutoWiringProtocolViolation(f"{source}")

    def _all_protocol_sources(self) -> Iterator[str]:
        def flatten(items: Iterable[Iterable[str]]) -> Iterable[str]:
            return chain.from_iterable(items)

        down_sources = flatten(self._protocol.down_sinks.values())
        up_sources = flatten(self._protocol.up_sinks.values())
        return chain(down_sources, up_sources)

    def _all_protocol_sinks(self) -> Iterator[str]:
        return chain(self._protocol.up_sinks, self._protocol.down_sinks)

    def connections(self) -> dict[tuple[str, str], tuple[str, str]]:
        return self._connections

    def wire(self, top: DataFlowNode, graph: Iterable[DataFlowNode]):
        self._check_protocol_support_violations(graph=graph, top=top)
        self._reset()
        self._remember_sources_provided_by(top)
        self._remember_unsatisfied_down_sinks_from(top)
        g = chain(graph, (top,))
        for node in g:
            self._connect_all_down_sinks_for(node)
            self._connect_all_up_sinks_for(node)
            self._remember_sources_provided_by(node)
            self._remember_unsatisfied_down_sinks_from(node)

    def _remember_unsatisfied_down_sinks_from(self, node: "DataFlowNode"):
        for s in self._protocol_down_sinks(node):
            if (node.name, s) not in self._unsatisfied_down_sinks:
                self._unsatisfied_down_sinks.append((node.name, s))

    def _connect_all_down_sinks_for(self, node: "DataFlowNode"):
        unsatisfied_down_sinks = self._unsatisfied_down_sinks.copy()
        for node_name, sink_name in unsatisfied_down_sinks:
            for source in self._protocol_down_sources(sink_name, node):
                self._connect_down_sink(
                    sink=(node_name, sink_name), source=(node.name, source)
                )
                break

    def _connect_all_up_sinks_for(self, node: "DataFlowNode"):
        for sink in self._protocol_up_sinks(node):
            self._connect_up_sink(node.name, sink)

    def _update_available_sources(self, matched_sink: str, node_name: str, source: str):
        self._available_sources[matched_sink] = (node_name, source)

    def _remember_sources_provided_by(self, node: "DataFlowNode"):
        for sink in self._protocol.up_sinks:
            sources = self._protocol_up_sources(sink, node)
            for source in sources:
                self._update_available_sources(sink, node.name, source)

    def _reset(self) -> None:
        self._available_sources: dict[str, tuple[str, str]] = {}
        self._unsatisfied_down_sinks: list[tuple[str, str]] = list()
        self._connections: dict[tuple[str, str], tuple[str, str]] = {}
        self._protocol = BASIC_WIRING

    @staticmethod
    def _intersection_iterator(
        a: Iterable[T], b: Sequence[T] | dict[T, Any]
    ) -> Iterator[T]:
        for s in a:
            if s in b:
                yield s

    def _protocol_down_sinks(self, node) -> Iterator[str]:
        yield from self._intersection_iterator(node.sinks, self._protocol.down_sinks)

    def _protocol_up_sinks(self, node) -> Iterator[str]:
        yield from self._intersection_iterator(node.sinks, self._protocol.up_sinks)

    def _protocol_up_sources(self, sink: str, node: "DataFlowNode") -> Iterator[str]:
        yield from self._intersection_iterator(
            node.sources, self._protocol.up_sinks[sink]
        )

    def _protocol_down_sources(self, sink: str, node: "DataFlowNode") -> Iterator[str]:
        yield from self._intersection_iterator(
            node.sources, self._protocol.down_sinks[sink]
        )

    def _connect_up_sink(self, node_name: str, sink_name: str):
        self._connections[(node_name, sink_name)] = self._available_sources[sink_name]

    def _connect_down_sink(self, source: tuple[str, str], sink: tuple[str, str]):
        self._connections[sink] = source
        self._unsatisfied_down_sinks.remove(sink)



from elasticai.creator.base_modules.hard_tanh import HardTanh as HardTanhBase
from elasticai.creator.nn.fixed_point._two_complement_fixed_point_config import (
    FixedPointConfig,
)
from elasticai.creator.vhdl.design_creator import DesignCreator

from .design import HardTanh as HardTanhDesign


class HardTanh(DesignCreator, HardTanhBase):
    def __init__(
        self, total_bits: int, frac_bits: int, min_val: float = -1, max_val: float = 1
    ) -> None:
        super().__init__(min_val, max_val)
        self._config = FixedPointConfig(total_bits=total_bits, frac_bits=frac_bits)

    def create_design(self, name: str) -> HardTanhDesign:
        return HardTanhDesign(
            name=name,
            total_bits=self._config.total_bits,
            frac_bits=self._config.frac_bits,
            min_val=self._config.as_integer(self.min_val),
            max_val=self._config.as_integer(self.max_val),
        )



from itertools import chain

from elasticai.creator.file_generation.savable import Path
from elasticai.creator.file_generation.template import (
    InProjectTemplate,
    module_to_package,
)
from elasticai.creator.vhdl.auto_wire_protocols.port_definitions import create_port
from elasticai.creator.vhdl.design.design import Design
from elasticai.creator.vhdl.design.ports import Port
from elasticai.creator.vhdl.shared_designs.rom import Rom


class Linear(Design):
    def __init__(
        self,
        *,
        in_feature_num: int,
        out_feature_num: int,
        total_bits: int,
        frac_bits: int,
        weights: list[list[int]],
        bias: list[int],
        name: str,
        work_library_name: str = "work",
        resource_option: str = "auto",
    ) -> None:
        super().__init__(name=name)
        self.weights = weights
        self.bias = bias
        self.in_feature_num = in_feature_num
        self.out_feature_num = out_feature_num
        self.work_library_name = work_library_name
        self.resource_option = resource_option
        self.frac_width = frac_bits
        self.data_width = total_bits
        self.x_addr_width = self.port["x_address"].width
        self.y_addr_width = self.port["y_address"].width

    @property
    def port(self) -> Port:
        return create_port(
            x_width=self.data_width,
            y_width=self.data_width,
            x_count=self.in_feature_num,
            y_count=self.out_feature_num,
        )

    def _template_parameters(self) -> dict[str, str]:
        return dict(
            (key, str(getattr(self, key)))
            for key in (
                "data_width",
                "frac_width",
                "x_addr_width",
                "y_addr_width",
                "in_feature_num",
                "out_feature_num",
            )
        )

    @staticmethod
    def _flatten_params(params: list[list[int]]) -> list[int]:
        return list(chain(*params))

    def save_to(self, destination: Path):
        rom_name = dict(weights=f"{self.name}_w_rom", bias=f"{self.name}_b_rom")

        template = InProjectTemplate(
            package=module_to_package(self.__module__),
            file_name="linear.tpl.vhd",
            parameters=dict(
                layer_name=self.name,
                weights_rom_name=rom_name["weights"],
                bias_rom_name=rom_name["bias"],
                work_library_name=self.work_library_name,
                resource_option=f'"{self.resource_option}"',
                **self._template_parameters(),
            ),
        )
        destination.create_subpath(self.name).as_file(".vhd").write(template)

        weights_rom = Rom(
            name=rom_name["weights"],
            data_width=self.data_width,
            values_as_integers=self._flatten_params(self.weights),
        )
        weights_rom.save_to(destination.create_subpath(rom_name["weights"]))

        bias_rom = Rom(
            name=rom_name["bias"],
            data_width=self.data_width,
            values_as_integers=self.bias,
        )
        bias_rom.save_to(destination.create_subpath(rom_name["bias"]))



import torch

from .conv1d import MathOperations as Conv1dOps
from .linear import MathOperations as LinearOps
from .lstm_cell import MathOperations as LSTMOps


class TorchMathOperations(LinearOps, Conv1dOps, LSTMOps):
    def quantize(self, a: torch.Tensor) -> torch.Tensor:
        return a

    def add(self, a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
        return a + b

    def mul(self, a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
        return a * b

    def matmul(self, a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
        return torch.matmul(a, b)



from .conv1d import BatchNormedConv1d, Conv1d
from .hard_sigmoid import HardSigmoid
from .hard_tanh import HardTanh
from .linear import BatchNormedLinear, Linear
from .precomputed import AdaptableSiLU, Sigmoid, Tanh
from .quantization import quantize
from .relu import ReLU



from elasticai.creator.base_modules.adaptable_silu import (
    AdaptableSiLU as AdaptableSiLUBase,
)

from .precomputed_module import PrecomputedModule


class AdaptableSiLU(PrecomputedModule):
    def __init__(
        self,
        total_bits: int,
        frac_bits: int,
        num_steps: int,
        sampling_intervall: tuple[float, float] = (-10, 10),
    ) -> None:
        super().__init__(
            base_module=AdaptableSiLUBase(),
            total_bits=total_bits,
            frac_bits=frac_bits,
            num_steps=num_steps,
            sampling_intervall=sampling_intervall,
        )



from typing import cast

import torch

from elasticai.creator.base_modules.conv1d import MathOperations as Conv1dOps
from elasticai.creator.base_modules.linear import MathOperations as LinearOps
from elasticai.creator.base_modules.lstm_cell import MathOperations as LSTMOps

from ._binary_quantization_function import Binarize


class MathOperations(LinearOps, Conv1dOps, LSTMOps):
    def quantize(self, a: torch.Tensor) -> torch.Tensor:
        return cast(torch.Tensor, Binarize.apply(a))

    def add(self, a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
        return self.quantize(a + b)

    def mul(self, a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
        return self.quantize(a * b)

    def matmul(self, a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
        return self.quantize(torch.matmul(a, b))



from abc import abstractmethod
from typing import Protocol

from torch import Tensor


class Quantize(Protocol):
    @abstractmethod
    def quantize(self, a: Tensor) -> Tensor: ...


class MatMul(Protocol):
    @abstractmethod
    def matmul(self, a: Tensor, b: Tensor) -> Tensor: ...


class Add(Protocol):
    @abstractmethod
    def add(self, a: Tensor, b: Tensor) -> Tensor: ...


class Mul(Protocol):
    @abstractmethod
    def mul(self, a: Tensor, b: Tensor) -> Tensor: ...



from elasticai.creator.base_modules.relu import ReLU as ReLUBase
from elasticai.creator.vhdl.design_creator import DesignCreator

from .design import ReLU as ReLUDesign


class ReLU(DesignCreator, ReLUBase):
    def __init__(self, total_bits: int, use_clock: bool = False) -> None:
        super().__init__()
        self._total_bits = total_bits
        self._use_clock = use_clock

    def create_design(self, name: str) -> ReLUDesign:
        return ReLUDesign(
            name=name,
            total_bits=self._total_bits,
            use_clock=self._use_clock,
        )



from typing import Any, Protocol

from elasticai.creator.file_generation.on_disk_path import OnDiskPath
from elasticai.creator.file_generation.savable import Path
from elasticai.creator.vhdl.ghdl_simulation import GHDLSimulator


class Design(Protocol):
    def save_to(self, destination: Path): ...


class Layer(Design, Protocol):
    @property
    def name(self) -> str: ...

    def create_testbench(self, name: str) -> "TestBench": ...


class TestBench(Design, Protocol):
    def set_inputs(self, *inputs) -> None: ...

    def parse_reported_content(self, content: Any) -> Any: ...


class SimulationConstructor(Protocol):
    def __call__(self, workdir: str, top_design_name: str) -> GHDLSimulator: ...


class SimulatedLayer:
    def __init__(
        self,
        layer_under_test: Layer,
        simulation_constructor: SimulationConstructor,
        working_dir: str,
    ):
        self._layer_under_test = layer_under_test
        self._simulation_constructor = simulation_constructor
        self._working_dir = working_dir
        self._inputs = None

    def __call__(self, *inputs):
        root = OnDiskPath("main", parent=self._working_dir)
        testbench_name = f"testbench_{self._layer_under_test.name}"
        testbench = self._layer_under_test.create_testbench(testbench_name)
        testbench.set_inputs(*inputs)
        testbench.save_to(root)
        runner = self._simulation_constructor(
            workdir=f"{self._working_dir}", top_design_name=testbench.name
        )
        runner.initialize()
        runner.run()
        actual = testbench.parse_reported_content(runner.getReportedContent())
        return actual



from typing import Any

import torch


class RoundToFloat(torch.autograd.Function):
    @staticmethod
    def forward(ctx: Any, *args: Any, **kwargs: Any) -> torch.Tensor:
        if len(args) != 3:
            raise TypeError(
                "apply() takes exactly three arguments "
                "(x: torch.Tensor, mantissa_bits: int, exponent_bits: int)"
            )
        x: torch.Tensor = args[0]
        mantissa_bits: int = args[1]
        exponent_bits: int = args[2]

        exponent_bias = 2 ** (exponent_bits - 1)
        largest_value = (2 - 1 / 2**mantissa_bits) * 2 ** (
            2**exponent_bits - exponent_bias - 1
        )

        out_of_bounds = (x < -largest_value) | (x > largest_value)
        if torch.any(out_of_bounds):
            raise ValueError("Cannot quantize tensor. Values out of bounds.")

        smallest_value = 2 ** (1 - exponent_bias - mantissa_bits)
        x[(x > -smallest_value) & (x < smallest_value)] = smallest_value

        scale = 2 ** (x.abs().log2().floor() - mantissa_bits)
        return scale * torch.round(x / scale)

    @staticmethod
    def backward(ctx: Any, *grad_outputs: Any) -> Any:
        return *grad_outputs, None, None



from abc import abstractmethod
from typing import Protocol

from .template import Template


class File(Protocol):
    @abstractmethod
    def write(self, template: Template) -> None: ...


class Path(Protocol):
    @abstractmethod
    def as_file(self, suffix: str) -> File: ...

    @abstractmethod
    def create_subpath(self, subpath_name: str) -> "Path": ...


class Savable(Protocol):
    @abstractmethod
    def save_to(self, destination: Path) -> None: ...



"""
Here we collect several functions to convert fixed point,
integer and natural numbers to bit patterns and vice versa.

IMPORTANT: We assume, the numbers to be representable in
the target format!
"""


def _toggle_bits(number: int, total_bits: int) -> int:
    def invert(value: int) -> int:
        return value ^ int("1" * total_bits, 2)

    def discard_leading_bits(value: int) -> int:
        return value & int("1" * total_bits, 2)

    return discard_leading_bits(invert(abs(number)))


def _twos_complement(number, total_bits):
    return _toggle_bits(number, total_bits) + 1


def bits_to_integer(pattern: str) -> int:
    is_negative = pattern[0] == "1"
    number = int(pattern, 2)
    if is_negative:
        number = -_twos_complement(number, len(pattern))
    return number


def bits_to_rational(pattern: str, frac_bits: int) -> float:
    pattern = pattern.strip()
    number = bits_to_integer(pattern)
    return number / (1 << frac_bits)


def convert_rational_to_bit_pattern(
    rational: float, total_bits: int, frac_bits: int
) -> str:
    return integer_to_bits(int(rational * (1 << frac_bits)), total_bits=total_bits)


def bits_to_natural(pattern: str) -> int:
    return int(pattern, 2)


def integer_to_bits(number: int, total_bits: int) -> str:
    if number < 0:
        number = _twos_complement(number, total_bits)
    return f"{number:0{total_bits}b}"


def _max_twos_complement_pattern(total_bits):
    return "0" + "1" * (total_bits - 1)


def _min_twos_complement_pattern(total_bits):
    return "1" + "0" * (total_bits - 1)


def max_rational(total_bits: int, frac_bits: int) -> float:
    return bits_to_rational(
        _max_twos_complement_pattern(total_bits), frac_bits=frac_bits
    )


def min_rational(total_bits: int, frac_bits: int) -> float:
    return bits_to_rational(
        _min_twos_complement_pattern(total_bits), frac_bits=frac_bits
    )


def min_integer(total_bits: int) -> int:
    return bits_to_integer(_min_twos_complement_pattern(total_bits))


def max_integer(total_bits: int) -> int:
    return bits_to_integer(_max_twos_complement_pattern(total_bits))


def min_natural(total_bits: int) -> int:
    return 0


def max_natural(total_bits: int) -> int:
    return bits_to_natural("1" * total_bits)



import os
from pathlib import Path as _PyPath

from .savable import File, Path
from .template import Template, TemplateExpander


class OnDiskFile(File):
    def __init__(self, full_path: str) -> None:
        self._full_path = full_path

    def write(self, template: Template) -> None:
        expander = TemplateExpander(template)
        unfilled_variables = expander.unfilled_variables()
        if len(unfilled_variables) > 0:
            raise KeyError(
                "Template is not filled completly. The following variables are"
                f" unfilled: {', '.join(unfilled_variables)}."
            )
        full_path = _PyPath(self._full_path)
        folder = full_path.parent
        if not folder.exists():
            os.makedirs(folder)
        with open(full_path, "w") as f:
            f.writelines((f"{line}\n" for line in expander.lines()))


class OnDiskPath(Path):
    def __init__(self, name: str, parent: str = ".") -> None:
        self._full_path = f"{parent}/{name}"

    def create_subpath(self, name: str) -> "OnDiskPath":
        return OnDiskPath(name, parent=self._full_path)

    def as_file(self, suffix: str) -> OnDiskFile:
        return OnDiskFile(full_path=f"{self._full_path}{suffix}")



from elasticai.creator.file_generation.savable import Path, Savable
from elasticai.creator.file_generation.template import InProjectTemplate


class MacDesign(Savable):
    def __init__(self, name: str, vector_width: int, fxp_params):
        self._name = name
        self._vector_width = vector_width
        self._fxp_params = fxp_params

    def save_to(self, destination: Path) -> None:
        wrapper = InProjectTemplate(
            package="elasticai.creator.nn.fixed_point.mac",
            file_name="mac_design.tpl.vhd",
            parameters={
                "total_width": str(self._fxp_params.total_bits),
                "frac_width": str(self._fxp_params.frac_bits),
                "vector_width": str(self._vector_width),
                "name": self._name,
            },
        )
        core_component = InProjectTemplate(
            package="elasticai.creator.nn.fixed_point.mac",
            file_name="fxp_mac.tpl.vhd",
            parameters={},
        )
        destination.create_subpath("fxp_mac").as_file(".vhd").write(core_component)
        destination.create_subpath(self._name).as_file(".vhd").write(wrapper)



from typing import cast

import torch
from torch import Tensor

from elasticai.creator.base_modules.conv1d import MathOperations as Conv1dOps
from elasticai.creator.base_modules.linear import MathOperations as LinearOps
from elasticai.creator.base_modules.lstm_cell import MathOperations as LSTMOps

from ._round_to_fixed_point import RoundToFixedPoint
from ._two_complement_fixed_point_config import FixedPointConfig


class MathOperations(LinearOps, Conv1dOps, LSTMOps):
    def __init__(self, config: FixedPointConfig) -> None:
        self.config = config

    def quantize(self, a: torch.Tensor) -> torch.Tensor:
        return self._round(self._clamp(a))

    def _clamp(self, a: torch.Tensor) -> torch.Tensor:
        return torch.clamp(
            a, min=self.config.minimum_as_rational, max=self.config.maximum_as_rational
        )

    def _round(self, a: torch.Tensor) -> torch.Tensor:
        return cast(torch.Tensor, RoundToFixedPoint.apply(a, self.config))

    def add(self, a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
        return self._clamp(a + b)

    def matmul(self, a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
        return self.quantize(torch.matmul(a, b))

    def mul(self, a: Tensor, b: Tensor) -> Tensor:
        return self.quantize(a * b)



from elasticai.creator.file_generation.savable import Path
from elasticai.creator.file_generation.template import InProjectTemplate

from ..number_converter import NumberConverter


class MacTestBench:
    def __init__(self, uut, fxp_params, name, uut_name):
        self._fxp_params = fxp_params
        self._converter = NumberConverter(fxp_params)
        self._uut = uut
        self._uut_name = uut_name
        self._inputs = None
        self._destination = None
        self.name = name

    def set_inputs(self, *inputs):
        self._inputs = inputs

    def parse_reported_content(self, outputs: list[str]):
        return self._converter.bits_to_rational(outputs[0])

    @property
    def _vector_width(self) -> str:
        return str(len(self._inputs[0]))

    @property
    def _total_bits(self) -> str:
        return str(self._fxp_params.total_bits)

    @property
    def _frac_bits(self) -> str:
        return str(self._fxp_params.frac_bits)

    def save_to(self, destination: Path):
        self._destination = destination
        inputs = self._prepare_inputs_for_test_bench(self._inputs)
        test_bench = InProjectTemplate(
            package="elasticai.creator.nn.fixed_point.mac",
            file_name="testbench.tpl.vhd",
            parameters=inputs
            | {
                "total_width": self._total_bits,
                "vector_width": str(self._vector_width),
                "uut_name": self._uut_name,
            },
        )
        self._uut.save_to(destination)
        destination.create_subpath(self.name).as_file(".vhd").write(test_bench)

    def _prepare_inputs_for_test_bench(self, inputs):
        x1, x2 = inputs
        x1 = map(self._converter.rational_to_bits, x1)
        x2 = map(self._converter.rational_to_bits, x2)
        inputs = {
            "x1": ", ".join([f'b"{x}"' for x in x1]),
            "x2": ", ".join([f'b"{x}"' for x in x2]),
        }
        return inputs



from abc import ABC, abstractmethod

from elasticai.creator.file_generation.savable import Path, Savable
from elasticai.creator.vhdl.design.ports import Port


class Design(Savable, ABC):
    def __init__(self, name: str):
        self.name = name

    @property
    @abstractmethod
    def port(self) -> Port: ...

    @abstractmethod
    def save_to(self, destination: Path) -> None: ...



from typing import cast

import torch

from elasticai.creator.nn.fixed_point._math_operations import MathOperations
from elasticai.creator.nn.fixed_point._two_complement_fixed_point_config import (
    FixedPointConfig,
)
from elasticai.creator.vhdl.design_creator import DesignCreator
from elasticai.creator.vhdl.shared_designs.precomputed_scalar_function import (
    PrecomputedScalarFunction,
)

from .identity_step_function import IdentityStepFunction


class PrecomputedModule(torch.nn.Module, DesignCreator):
    def __init__(
        self,
        base_module: torch.nn.Module,
        total_bits: int,
        frac_bits: int,
        num_steps: int,
        sampling_intervall: tuple[float, float],
    ) -> None:
        super().__init__()
        self._base_module = base_module
        self._config = FixedPointConfig(total_bits=total_bits, frac_bits=frac_bits)
        self._operations = MathOperations(self._config)
        self._step_lut = torch.nn.Parameter(
            torch.linspace(*sampling_intervall, num_steps),
            requires_grad=False,
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self._stepped_inputs(x)
        outputs = self._base_module(x)
        return self._operations.quantize(outputs)

    def create_design(self, name: str) -> PrecomputedScalarFunction:
        quantized_inputs = list(map(self._config.as_integer, self._step_lut.tolist()))
        return PrecomputedScalarFunction(
            name=name,
            width=self._config.total_bits,
            inputs=quantized_inputs,
            function=self._quantized_inference,
        )

    def _stepped_inputs(self, x: torch.Tensor) -> torch.Tensor:
        step_inputs = cast(torch.Tensor, IdentityStepFunction.apply(x, self._step_lut))
        return self._operations.quantize(step_inputs)

    def _quantized_inference(self, x: int) -> int:
        fxp_input = self._config.as_rational(x)
        with torch.no_grad():
            output = self.cpu()(torch.tensor(fxp_input))
        return self._config.as_integer(float(output.item()))



from typing import Optional, cast

from elasticai.creator.vhdl.code_generation.addressable import calculate_address_width
from elasticai.creator.vhdl.code_generation.vhdl_ports import (
    template_string_for_port_definition,
)
from elasticai.creator.vhdl.design import std_signals as _signals
from elasticai.creator.vhdl.design.ports import Port


def create_port(
    x_width: int,
    y_width: int,
    *,
    x_count: int = 0,
    y_count: int = 0,
    x_address_width: Optional[int] = None,
    y_address_width: Optional[int] = None,
) -> Port:
    if (
        x_count == 0
        and y_count == 0
        and x_address_width is None
        and y_address_width is None
    ):
        return create_port_for_bufferless_design(x_width, y_width)
    else:
        return create_port_for_buffered_design(
            x_width=x_width,
            y_width=y_width,
            x_count=x_count,
            y_count=y_count,
            x_address_width=x_address_width,
            y_address_width=y_address_width,
        )


def create_port_for_bufferless_design(x_width: int, y_width: int) -> Port:
    in_signals = [
        _signals.enable(),
        _signals.clock(),
        _signals.x(x_width),
    ]
    out_signals = [
        _signals.y(y_width),
    ]
    return Port(incoming=in_signals, outgoing=out_signals)


def create_port_for_buffered_design(
    *,
    x_width: int,
    y_width: int,
    x_count: int,
    y_count: int,
    x_address_width: Optional[int] = None,
    y_address_width: Optional[int] = None,
) -> Port:
    if not (
        (not (x_address_width is None and y_address_width is None))
        or (x_count > 0 or y_count > 0)
    ):
        raise ValueError(
            "Provide either (x_count, y_count) or (x_address_width, y_address_width)"
        )
    if x_address_width is None and y_address_width is None:
        x_address_width = calculate_address_width(x_count)
        y_address_width = calculate_address_width(y_count)

    in_signals = [
        _signals.enable(),
        _signals.clock(),
        _signals.x(x_width),
        _signals.y_address(cast(int, y_address_width)),
    ]
    out_signals = [
        _signals.done(),
        _signals.y(y_width),
        _signals.x_address(cast(int, x_address_width)),
    ]
    return Port(incoming=in_signals, outgoing=out_signals)


def port_definition_template_for_buffered_design() -> list[str]:
    return template_string_for_port_definition(
        create_port_for_buffered_design(x_width=1, y_width=1, x_count=1, y_count=1)
    )


def port_definition_template_for_bufferless_design() -> list[str]:
    return template_string_for_port_definition(create_port_for_bufferless_design(1, 1))



from typing import Any, cast

import torch

from elasticai.creator.base_modules.conv1d import Conv1d as Conv1dBase
from elasticai.creator.nn.fixed_point._math_operations import MathOperations
from elasticai.creator.nn.fixed_point._two_complement_fixed_point_config import (
    FixedPointConfig,
)
from elasticai.creator.nn.fixed_point.conv1d.design import Conv1d as Conv1dDesign
from elasticai.creator.vhdl.design_creator import DesignCreator


class Conv1d(DesignCreator, Conv1dBase):
    def __init__(
        self,
        total_bits: int,
        frac_bits: int,
        in_channels: int,
        out_channels: int,
        signal_length: int,
        kernel_size: int | tuple[int],
        bias: bool = True,
        device: Any = None,
    ) -> None:
        self._config = FixedPointConfig(total_bits=total_bits, frac_bits=frac_bits)
        self._signal_length = signal_length
        super().__init__(
            operations=MathOperations(config=self._config),
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            bias=bias,
            device=device,
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        has_batches = x.dim() == 2

        input_shape = (
            (x.shape[0], self.in_channels, -1)
            if has_batches
            else (self.in_channels, -1)
        )
        output_shape = (x.shape[0], -1) if has_batches else (-1,)

        x = x.view(*input_shape)
        outputs = super().forward(x)
        return outputs.view(*output_shape)

    def create_design(self, name: str) -> Conv1dDesign:
        def float_to_signed_int(value: float | list) -> int | list:
            if isinstance(value, list):
                return list(map(float_to_signed_int, value))
            return self._config.as_integer(value)

        def flatten_tuple(x: int | tuple[int, ...]) -> int:
            return x[0] if isinstance(x, tuple) else x

        bias = [0] * self.out_channels if self.bias is None else self.bias.tolist()
        signed_int_weights = cast(
            list[list[list[int]]], float_to_signed_int(self.weight.tolist())
        )
        signed_int_bias = cast(list[int], float_to_signed_int(bias))

        return Conv1dDesign(
            name=name,
            total_bits=self._config.total_bits,
            frac_bits=self._config.frac_bits,
            in_channels=self.in_channels,
            out_channels=self.out_channels,
            signal_length=self._signal_length,
            kernel_size=flatten_tuple(self.kernel_size),
            weights=signed_int_weights,
            bias=signed_int_bias,
        )



from elasticai.creator.file_generation.savable import Savable
from elasticai.creator.vhdl.simulated_layer import SimulatedLayer

from .._math_operations import MathOperations
from .._two_complement_fixed_point_config import FixedPointConfig
from ..number_converter import FXPParams
from .design import MacDesign
from .mactestbench import MacTestBench


class MacLayer:
    def __init__(self, vector_width: int, fxp_params: FXPParams):
        self.ops = MathOperations(
            FixedPointConfig(
                total_bits=fxp_params.total_bits, frac_bits=fxp_params.frac_bits
            )
        )
        self._vector_width = vector_width
        self._fxp_params = fxp_params
        self.name = "fxp_mac"

    def __call__(self, a, b):
        return self.ops.matmul(a, b)

    def create_design(self, name: str) -> Savable:
        return MacDesign(
            fxp_params=self._fxp_params, vector_width=self._vector_width, name=name
        )

    def create_testbench(self, name: str) -> MacTestBench:
        return MacTestBench(
            fxp_params=self._fxp_params,
            uut=self.create_design("mac_wrapper_test"),
            uut_name="mac_wrapper_test",
            name=name,
        )

    def create_simulation(self, simulator, working_dir):
        return SimulatedLayer(self, simulator, working_dir=working_dir)
